/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef bp_h
#define bp_h

#include <bp/bp_TA_type.h>
#include <pdp/base.h>
#include <pdp/netstru.h>
#include <pdp/sched_proc.h>

class BpConSpec;
class BpCon;
class BpCon_Group;
class BpUnit;
class BpUnitSpec;
class BpTrial;


class BpCon : public Connection {
  // Bp connection
public:
  float 		dwt; 		// #NO_SAVE Change in weight
  float 		dEdW; 		// #NO_SAVE derivative of Error wrt weight

  void 	Initialize()		{ dwt = dEdW = 0.0f; }
  void 	Destroy()		{ };
  void	Copy_(const BpCon& cp)	{ dwt = cp.dwt; dEdW = cp.dEdW; }
  COPY_FUNS(BpCon, Connection);
  TA_BASEFUNS(BpCon);
};

// ConSpec now has 3 versions of some functions, the two regular ones
// (one for the group and one for the connection), and now a third which is a
// virtual version of the C_ for use with the bias weights.
// every time a C_ version is overloaded, it is necessary to overload the B_
// version, where the new B_ just calls the new C_

class BpConSpec : public ConSpec {
  // Bp connection specifications
public:
  enum MomentumType {
    AFTER_LRATE,		// apply momentum after learning rate (old pdp style)
    BEFORE_LRATE,		// apply momentum before learning rate
    NORMALIZED 			// like BEFORE, but normalize direction to unit length
  };

  float 	lrate;		// learning rate
  float 	momentum;	// momentum
  MomentumType	momentum_type;	// type of momentum function to use
  float		momentum_c;	// #READ_ONLY complement of momentum (for NORMALIZED)
  float 	decay;		// decay rate (before lrate and momentum)
  void 		(*decay_fun)(BpConSpec* spec, BpCon* cn, BpUnit* ru, BpUnit* su);
  // #LIST_BpConSpec_WtDecay the weight decay function to use

  void 		C_InitWtDelta(Con_Group* cg, Connection* cn, Unit* ru, Unit* su)
  { ConSpec::C_InitWtDelta(cg, cn, ru, su); ((BpCon*)cn)->dEdW = 0.0f; }

  void 		C_InitWtState(Con_Group* cg, Connection* cn, Unit* ru, Unit* su) 	
  { ConSpec::C_InitWtState(cg, cn, ru, su); ((BpCon*)cn)->dwt = 0.0f;}

  inline float		C_Compute_dEdA(BpCon* cn, BpUnit* ru, BpUnit* su);
  inline virtual float 	Compute_dEdA(BpCon_Group* cg, BpUnit* su);
  // get error from units I send to

  inline void 		C_Compute_dWt(BpCon* cn, BpUnit* ru, BpUnit* su);
  inline void 		Compute_dWt(Con_Group* cg, Unit* ru);
  inline virtual void	B_Compute_dWt(BpCon* cn, BpUnit* ru);
  // Compute dE with respect to the weights

  inline void 	C_Compute_WtDecay(BpCon* cn, BpUnit* ru, BpUnit* su);
  // call the decay function 

  inline void	C_BEF_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su); // BEFORE_LRATE
  inline void	C_AFT_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su); // AFTER_LRATE
  inline void	C_NRM_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su); // NORMALIZED
  inline void	UpdateWeights(Con_Group* cg, Unit* ru);
  inline virtual void	B_UpdateWeights(BpCon* cn, BpUnit* ru);
  // for the bias unit

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(BpConSpec);
  COPY_FUNS(BpConSpec, ConSpec);
  TA_BASEFUNS(BpConSpec);
};

// the following functions are possible weight decay functions

// #REG_FUN
void Bp_Simple_WtDecay(BpConSpec* spec, BpCon* cn, BpUnit* ru, BpUnit* su) 
// #LIST_BpConSpec_WtDecay Simple weight decay (subtract decay*wt)
     ;				// term here so scanner picks up comment
// #REG_FUN
void Bp_WtElim_WtDecay(BpConSpec* spec, BpCon* cn, BpUnit* ru, BpUnit* su) 
// #LIST_BpConSpec_WtDecay Weight Elimination (Rumelhart) weight decay
     ;				// term here so scanner picks up comment

class BpCon_Group : public Con_Group {
  // group of Bp connections
public:
  // these are "convenience" functions for those defined in the spec

  float Compute_dEdA(BpUnit* su) { return ((BpConSpec*)spec.spec)->Compute_dEdA(this, su); }

  void	Initialize();
  void 	Destroy()		{ };
  TA_BASEFUNS(BpCon_Group);
};


class BpUnit;

class BpUnitSpec : public UnitSpec {
  // specifications for Bp units
public:
  SigmoidSpec	sig;		// sigmoid activation parameters
  float		err_tol;	// error tolerance (no error signal if |t-o|<err_tol)
  void 		(*err_fun)(BpUnitSpec* spec, BpUnit* u);
  // #LIST_BpUnit_Error this points to the error fun, set appropriately

  // generic unit functions
  void		InitState(Unit* u);
  void 		Compute_Act(Unit* u);		// activation from net input (sigmoid)
  void 		Compute_dWt(Unit* u); 		// for all of my recv weights
  void 		UpdateWeights(Unit* u);		// modify to update bias weight
  
  // bp special functions
  virtual void 	Compute_Error(BpUnit* u); 	// call the error function (testing only)
  virtual void 	Compute_dEdA(BpUnit* u); 	// error wrt unit activation
  virtual void 	Compute_dEdNet(BpUnit* u); 	// error wrt net input

  void 	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void	InitLinks();
  void	CutLinks();
  void	Copy_(const BpUnitSpec& cp);
  COPY_FUNS(BpUnitSpec, UnitSpec);
  TA_BASEFUNS(BpUnitSpec);
};

// the following functions are possible error functions. 

// #REG_FUN
void Bp_Squared_Error(BpUnitSpec* spec, BpUnit* u) 
// #LIST_BpUnit_Error Squared error function for bp
     ;				// term here so scanner picks up comment
// #REG_FUN
void Bp_CrossEnt_Error(BpUnitSpec* spec, BpUnit* u)
// #LIST_BpUnit_Error Cross entropy error function for bp
     ;				// term here so scanner picks up comment


class BpUnit : public Unit {
  // standard feed-forward Bp unit
public:
  float 	err; 		// this is E, not dEdA
  float 	dEdA;		// #LABEL_dEdA error wrt activation
  float 	dEdNet;		// #LABEL_dEdNet error wrt net input

  // these are "convenience" functions for those defined in the spec
  void Compute_Error()		{ ((BpUnitSpec*)spec.spec)->Compute_Error(this); }
  void Compute_dEdA()		{ ((BpUnitSpec*)spec.spec)->Compute_dEdA(this); }
  void Compute_dEdNet()		{ ((BpUnitSpec*)spec.spec)->Compute_dEdNet(this); }
  void Compute_dEdA_dEdNet() 	{ Compute_dEdA(); Compute_dEdNet(); }

  void 	Initialize();
  void 	Destroy()		{ };
  void	Copy_(const BpUnit& cp);
  COPY_FUNS(BpUnit, Unit);
  TA_BASEFUNS(BpUnit);
};

// inline functions (for speed)

inline float BpConSpec::C_Compute_dEdA(BpCon* cn, BpUnit* ru, BpUnit*) {
  return cn->wt * ru->dEdNet;
}
inline float BpConSpec::Compute_dEdA(BpCon_Group* cg, BpUnit* su) {
  float rval = 0.0f;
  CON_GROUP_LOOP(cg,rval += C_Compute_dEdA((BpCon*)cg->Cn(i), (BpUnit*)cg->Un(i), su));
  return rval;
}


inline void BpConSpec::C_Compute_dWt(BpCon* cn, BpUnit* ru, BpUnit* su) {
  cn->dEdW += su->act * ru->dEdNet;
}
inline void BpConSpec::Compute_dWt(Con_Group* cg, Unit* ru) {
  CON_GROUP_LOOP(cg,C_Compute_dWt((BpCon*)cg->Cn(i), (BpUnit*)ru, (BpUnit*)cg->Un(i)));
}
inline void BpConSpec::B_Compute_dWt(BpCon* cn, BpUnit* ru) {
  cn->dEdW += ru->dEdNet;
}

inline void BpConSpec::C_Compute_WtDecay(BpCon* cn, BpUnit* ru, BpUnit* su) {
  if(decay_fun != NULL)
    (*decay_fun)(this, cn, ru, su);
}
inline void BpConSpec::C_AFT_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su) {
  C_Compute_WtDecay(cn, ru, su); // decay goes into dEdW...
  cn->dwt = lrate * cn->dEdW + momentum * cn->dwt;
  cn->wt += cn->dwt;
  cn->dEdW = 0.0f;
}
inline void BpConSpec::C_BEF_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su) {
  C_Compute_WtDecay(cn, ru, su);
  cn->dwt = cn->dEdW + momentum * cn->dwt;
  cn->wt += lrate * cn->dwt;
  cn->dEdW = 0.0f;
}
inline void BpConSpec::C_NRM_UpdateWeights(BpCon* cn, BpUnit* ru, BpUnit* su) {
  C_Compute_WtDecay(cn, ru, su);
  cn->dwt = momentum_c * cn->dEdW + momentum * cn->dwt;
  cn->wt += lrate * cn->dwt;
  cn->dEdW = 0.0f;
}

inline void BpConSpec::UpdateWeights(Con_Group* cg, Unit* ru) {
  if(momentum_type == AFTER_LRATE) {
    CON_GROUP_LOOP(cg, C_AFT_UpdateWeights((BpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  else if(momentum_type == BEFORE_LRATE) {
    CON_GROUP_LOOP(cg, C_BEF_UpdateWeights((BpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  else {
    CON_GROUP_LOOP(cg, C_NRM_UpdateWeights((BpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  ApplyLimits(cg, ru);
}

inline void BpConSpec::B_UpdateWeights(BpCon* cn, BpUnit* ru) {
  if(momentum_type == AFTER_LRATE) 
    C_AFT_UpdateWeights(cn, ru, NULL);
  else if(momentum_type == BEFORE_LRATE)
    C_BEF_UpdateWeights(cn, ru, NULL);
  else
    C_NRM_UpdateWeights(cn, ru, NULL);
  C_ApplyLimits(cn, ru, NULL);
}

class BpTrial : public TrialProcess {
  // standard Bp feed-forward trial
public:
  void 		Loop();
  bool 		Crit()		{ return true; } // executes loop only once

  virtual void	Compute_Act();	
  virtual void	Compute_Error();
  virtual void	Compute_dEdA_dEdNet();
  virtual void	Compute_dWt();

  void	Initialize();
  void 	Destroy()		{ };
  void	InitLinks();
  TA_BASEFUNS(BpTrial);
};

//////////////////////////////////////////
//	Additional Stat Types 		//
//////////////////////////////////////////

class CE_Stat : public Stat {
  // ##COMPUTE_IN_TrialProcess Cross-entropy error statistic (asymmetric divergence)
public:
  StatVal	ce;			// cross-entropy error
  float		tolerance;		// if error is less than this, its 0

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Unit_Stat(Unit* unit);

  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(CE_Stat);
  COPY_FUNS(CE_Stat, Stat);
  TA_BASEFUNS(CE_Stat);
};

class NormDotProd_Stat : public Stat {
  // ##COMPUTE_IN_TrialProcess Normalized Dot Product of act and target values
public:
  StatVal	ndp;		 // normalized dot product

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Unit_Stat(Unit* unit);
  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(NormDotProd_Stat);
  COPY_FUNS(NormDotProd_Stat, Stat);
  TA_BASEFUNS(NormDotProd_Stat);
};

class VecCor_Stat : public Stat {
  // ##COMPUTE_IN_TrialProcess Vector Correlation of act and target values
public:
  StatVal	vcor;		 // vector correlation
  float		dp;		 // #HIDDEN
  float		l1;		 // #HIDDEN
  float		l2;		 // #HIDDEN

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Unit_Stat(Unit* unit);
  void		Network_Stat();
  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(VecCor_Stat);
  COPY_FUNS(VecCor_Stat, Stat);
  TA_BASEFUNS(VecCor_Stat);
};

class NormVecLen_Stat : public Stat {
  // ##COMPUTE_IN_TrialProcess Normalized Vector Length of act and target values
public:
  StatVal	nvl;		 // normalized vector length

  void		RecvCon_Run(Unit*)	{ }; // don't do these!
  void		SendCon_Run(Unit*)	{ };

  void		InitStat();
  void		Init();
  bool		Crit();
  void		Network_Init();
  void		Layer_Run();		// only compute on TARG layers
  void 		Unit_Stat(Unit* unit);
  void		Network_Stat();
  void		NameStatVals();

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(NormVecLen_Stat);
  COPY_FUNS(NormVecLen_Stat, Stat);
  TA_BASEFUNS(NormVecLen_Stat);
};


//////////////////////////////////////////
//	Additional ConSpec Types	//
//////////////////////////////////////////

class HebbBpConSpec : public BpConSpec {
  // Simple Hebbian wt update (send act * recv act)
public:
  inline void 		C_Compute_dWt(BpCon* cn, BpUnit* ru, BpUnit* su);
  inline void 		Compute_dWt(Con_Group* cg, Unit* ru);

  inline void		B_Compute_dWt(BpCon* cn, BpUnit* ru);

  void	Initialize()		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(HebbBpConSpec);
};

inline void HebbBpConSpec::C_Compute_dWt(BpCon* cn, BpUnit* ru, BpUnit* su) {
  cn->dEdW += ((ru->ext_flag & Unit::TARG) ? ru->targ : ru->act) * su->act;
}

inline void HebbBpConSpec::Compute_dWt(Con_Group* cg, Unit* ru) {
  CON_GROUP_LOOP(cg,C_Compute_dWt((BpCon*)cg->Cn(i), (BpUnit*)ru, (BpUnit*)cg->Un(i)));
}

inline void HebbBpConSpec::B_Compute_dWt(BpCon* cn, BpUnit* ru) {
  cn->dEdW +=  ((ru->ext_flag & Unit::TARG) ? ru->targ : ru->act);
}

class ErrScaleBpConSpec : public BpConSpec {
  // con spec that scales the error by given parameter
public:
  float		err_scale;	// the scaling parameter

  inline float 		C_Compute_dEdA(BpCon* cn, BpUnit* ru, BpUnit* su);
  inline float 		Compute_dEdA(BpCon_Group* cg, BpUnit* su);

  void 	Initialize()	{ err_scale = 1.0f; }
  void 	Destroy()	{ };
  SIMPLE_COPY(ErrScaleBpConSpec);
  COPY_FUNS(ErrScaleBpConSpec, BpConSpec);
  TA_BASEFUNS(ErrScaleBpConSpec);
};

inline float ErrScaleBpConSpec::C_Compute_dEdA(BpCon* cn, BpUnit* ru, BpUnit*) {
  return err_scale * cn->wt * ru->dEdNet;
}
inline float ErrScaleBpConSpec::Compute_dEdA(BpCon_Group* cg, BpUnit* su) {
  float rval = 0.0f;
  CON_GROUP_LOOP(cg,rval += C_Compute_dEdA((BpCon*)cg->Cn(i), (BpUnit*)cg->Un(i), su));
  return rval;
}

class DeltaBarDeltaBpCon : public BpCon {
  // delta-bar-delta connection object with local learning rate
public:
  float 		lrate; 		// #NO_SAVE local learning rate

  void 	Initialize()		{ lrate = 0.0f; }
  void 	Destroy()		{ };
  void	Copy_(const DeltaBarDeltaBpCon& cp)	{ lrate = cp.lrate; }
  COPY_FUNS(DeltaBarDeltaBpCon, BpCon);
  TA_BASEFUNS(DeltaBarDeltaBpCon);
};

class DeltaBarDeltaBpConSpec : public BpConSpec {
  // delta-bar-delta has local learning rates for each connection
public:
  float		lrate_incr;	// rate of learning rate increase (additive)
  float		lrate_decr;	// rate of learning rate decrease (multiplicative)
  float		act_lrate_incr;	// #HIDDEN actual lrate increase (times lrate)

  void 		C_InitWtState(Con_Group* cg, Connection* cn, Unit* ru, Unit* su) 	
  { ConSpec::C_InitWtState(cg, cn, ru, su); ((DeltaBarDeltaBpCon*)cn)->lrate = lrate;}
  // set initial learning rate

  inline void	C_UpdateLrate(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su);
  inline void	C_BEF_UpdateWeights(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su);
  inline void	C_AFT_UpdateWeights(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su);
  inline void	C_NRM_UpdateWeights(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su);
  inline virtual void	UpdateWeights(Con_Group* cg, Unit* ru);
  inline virtual void	B_UpdateWeights(BpCon* cn, BpUnit* ru);

  void	UpdateAfterEdit();
  void	Initialize();
  void 	Destroy()		{ };
  void	Copy_(const DeltaBarDeltaBpConSpec& cp);
  COPY_FUNS(DeltaBarDeltaBpConSpec, BpConSpec);
  TA_BASEFUNS(DeltaBarDeltaBpConSpec);
};

inline void DeltaBarDeltaBpConSpec::C_UpdateLrate
(DeltaBarDeltaBpCon* cn, BpUnit*, BpUnit*)
{
  float prod = cn->dwt * cn->dEdW;
  if(prod > 0.0f)
    cn->lrate += act_lrate_incr;
  else if(prod < 0.0f)
    cn->lrate *= lrate_decr;
  // prod = 0 means first epoch, don't change lrate..
}

inline void DeltaBarDeltaBpConSpec::C_AFT_UpdateWeights
(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su)
{
  C_Compute_WtDecay(cn, ru, su); // decay goes into dEdW...
  C_UpdateLrate(cn, ru, su);
  cn->dwt = cn->lrate * cn->dEdW + momentum * cn->dwt;
  cn->wt += cn->dwt;
  cn->dEdW = 0.0f;
}
inline void DeltaBarDeltaBpConSpec::C_BEF_UpdateWeights
(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su)
{
  C_Compute_WtDecay(cn, ru, su);
  C_UpdateLrate(cn, ru, su);
  cn->dwt = cn->dEdW + momentum * cn->dwt;
  cn->wt += cn->lrate * cn->dwt;
  cn->dEdW = 0.0f;
}
inline void DeltaBarDeltaBpConSpec::C_NRM_UpdateWeights
(DeltaBarDeltaBpCon* cn, BpUnit* ru, BpUnit* su)
{
  C_Compute_WtDecay(cn, ru, su);
  C_UpdateLrate(cn, ru, su);
  cn->dwt = momentum_c * cn->dEdW + momentum * cn->dwt;
  cn->wt += cn->lrate * cn->dwt;
  cn->dEdW = 0.0f;
}

inline void DeltaBarDeltaBpConSpec::UpdateWeights(Con_Group* cg, Unit* ru) {
  if(momentum_type == AFTER_LRATE) {
    CON_GROUP_LOOP(cg, C_AFT_UpdateWeights((DeltaBarDeltaBpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  else if(momentum_type == BEFORE_LRATE) {
    CON_GROUP_LOOP(cg, C_BEF_UpdateWeights((DeltaBarDeltaBpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  else {
    CON_GROUP_LOOP(cg, C_NRM_UpdateWeights((DeltaBarDeltaBpCon*)cg->Cn(i),
					   (BpUnit*)ru, (BpUnit*)cg->Un(i)));
  }
  ApplyLimits(cg, ru);
}

inline void DeltaBarDeltaBpConSpec::B_UpdateWeights(BpCon* cn, BpUnit* ru) {
  if(momentum_type == AFTER_LRATE) 
    C_AFT_UpdateWeights((DeltaBarDeltaBpCon*)cn, ru, NULL);
  else if(momentum_type == BEFORE_LRATE)
    C_BEF_UpdateWeights((DeltaBarDeltaBpCon*)cn, ru, NULL);
  else
    C_NRM_UpdateWeights((DeltaBarDeltaBpCon*)cn, ru, NULL);
  C_ApplyLimits(cn, ru, NULL);
}


//////////////////////////////////////////
//	Additional Unit Types		//
//////////////////////////////////////////

class BpContextSpec : public BpUnitSpec {
  /* for context units in simple recurrent networks (SRN), expects a one-to-one prjn
     from the layer it copies from */
public:
  float		hysteresis;	 // hysteresis factor: (1-hyst)*new + hyst*old
  float		hysteresis_c;	 // #READ_ONLY complement of hysteresis
  Random	initial_act;	 // initial activation value
  String	variable;	 // name of unit variable to copy into
  Unit::ExtType	unit_flags;	 // flags to set on the unit after copying value
  MemberDef*	var_md;		 // #IGNORE memberdef of variable

  void		InitState(Unit* u);
  void 		Compute_Act(Unit* u);	
  // copy activation from corresponding unit in projection from layer

  // nullify all other functions..
  void 		Compute_Net(Unit*) 	{ };
  void 		InitWtDelta(Unit*) 	{ };
  void 		Compute_dWt(Unit*) 	{ }; 
  void 		UpdateWeights(Unit*) 	{ };
  
  // bp special functions
  void 	Compute_Error(BpUnit*)		{ };
  void 	Compute_dEdA(BpUnit*)		{ }; 	
  void 	Compute_dEdNet(BpUnit*)		{ };

  void	UpdateAfterEdit();
  void 	Initialize();
  void	Destroy()		{ };
  void	InitLinks();
  void	Copy_(const BpContextSpec& cp);
  COPY_FUNS(BpContextSpec, BpUnitSpec);
  TA_BASEFUNS(BpContextSpec);
};

class LinearBpUnitSpec : public BpUnitSpec {
  // linear unit in Bp
public:
  void 		Compute_Act(Unit* u);
  void		Compute_dEdNet(BpUnit* u);

  void	Initialize();
  void 	Destroy()		{ };
  TA_BASEFUNS(LinearBpUnitSpec);
};

class ThreshLinBpUnitSpec : public BpUnitSpec {
  // linear unit in Bp
public:
  float		threshold;

  void 		Compute_Act(Unit* u);
  void		Compute_dEdNet(BpUnit* u);

  void	Initialize();
  void 	Destroy()		{ };
  SIMPLE_COPY(ThreshLinBpUnitSpec);
  COPY_FUNS(ThreshLinBpUnitSpec, BpUnitSpec);
  TA_BASEFUNS(ThreshLinBpUnitSpec);
};

class NoisyBpUnitSpec : public BpUnitSpec {
  // Bp with noisy output signal (act plus noise)
public:
  Random	noise;		// what kind of noise to add to activations

  void 		Compute_Act(Unit* u);

  void	Initialize();
  void 	Destroy()		{ };
  void 	InitLinks();
  SIMPLE_COPY(NoisyBpUnitSpec);
  COPY_FUNS(NoisyBpUnitSpec, BpUnitSpec);
  TA_BASEFUNS(NoisyBpUnitSpec);
};

class StochasticBpUnitSpec : public BpUnitSpec {
  // Bp with a binary stochastic activation function
public:
  void				Compute_Act(Unit* u);
  void	Initialize()		{ };
  void 	Destroy()		{ };
  TA_BASEFUNS(StochasticBpUnitSpec);
};

class RBFBpUnitSpec : public BpUnitSpec {
  // Radial basis function (Gaussian) function units in Bp
public:
  float         var;            // variance of Gaussian
  float         norm_const;     // #HIDDEN normalization const for Gaussian
  float         denom_const;    // #HIDDEN denominator const for Gaussian

  void          Compute_Net(Unit* u);
  void          Compute_Act(Unit* u);
  void          Compute_dEdNet(BpUnit* u);

  void  UpdateAfterEdit();
  void  Initialize();
  void  Destroy()               { };
  SIMPLE_COPY(RBFBpUnitSpec);
  COPY_FUNS(RBFBpUnitSpec, BpUnitSpec);
  TA_BASEFUNS(RBFBpUnitSpec);
};

class BumpBpUnitSpec : public BpUnitSpec {
  // bump function in Bp: Gaussian of std net input
public:
  float         mean;           // mean of Gaussian
  float         std_dev;        // std deviation of Gaussian
  float         std_dev_r;      // #HIDDEN reciprocal of std_dev

  void          Compute_Act(Unit* u);
  void          Compute_dEdNet(BpUnit* u);

  void  UpdateAfterEdit();
  void  Initialize();
  void  Destroy()               { };
  SIMPLE_COPY(BumpBpUnitSpec);
  COPY_FUNS(BumpBpUnitSpec, BpUnitSpec);
  TA_BASEFUNS(BumpBpUnitSpec);
};

class ExpBpUnitSpec : public BpUnitSpec {
  // exponential units in Bp: simple exponent of net input
public:
  void          Compute_Act(Unit* u);
  virtual void  Compute_dEdNet(BpUnit* u);      // error wrt unit activation

  void  Initialize()	{ };
  void  Destroy()	{ };
  TA_BASEFUNS(ExpBpUnitSpec);
};

class SoftMaxBpUnitSpec : public BpUnitSpec {
  /* SoftMax Units: first one-to-one prjn is from corresp exponential unit,
     second prjn is from single summing linear unit, this then divides two */
public:
  
  void		Compute_Net(Unit*)	{ }; // do nothing
  void          Compute_Act(Unit* u);
  void          Compute_dEdNet(BpUnit* u);

  // don't update my weights
  void          Compute_dWt(Unit*)	{ };
  void          UpdateWeights(Unit*)	{ };

  void  Initialize()	{ };
  void  Destroy()	{ };
  TA_BASEFUNS(SoftMaxBpUnitSpec);
};

  
#endif // bp.h
