/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// stats.cc

#include <pdp/stats.h>
#include <pdp/netstru.h>
#include <pdp/sched_proc.h>

#include <values.h>

//////////////////////////
//  	CritParam      	//
//////////////////////////

void CritParam::Initialize() {
  flag = false;
  rel = LESSTHANOREQUAL;
  val = 0.0f;
  n_met = 0;
  cnt = 1;
}; 

void CritParam::Copy_(const CritParam& cp) {
  flag = cp.flag;
  rel = cp.rel;
  val = cp.val;
  cnt = cp.cnt;
  n_met = cp.n_met;
}

bool CritParam::Evaluate(float cmp) {
  if(!flag) return false;
  bool met = false;
  switch(rel) {
  case EQUAL:
    if(cmp == val)	met = true;
    break;
  case NOTEQUAL:         
    if(cmp != val)	met = true;
    break;
  case LESSTHAN:         
    if(cmp < val)	met = true;
    break;
  case GREATERTHAN:      
    if(cmp > val)	met = true;
    break;
  case LESSTHANOREQUAL:  
    if(cmp <= val)	met = true;
    break;
  case GREATERTHANOREQUAL:
    if(cmp >= val)	met = true;
    break;
  }
  if(met) {
    n_met++;
    if(n_met >= cnt) return true;
  }
  else
    n_met = 0;		// reset the counter (has to be consecutive)
  return false;
}


//////////////////////////
//  	StatVal     	//
//////////////////////////

void StatVal::Initialize() {
  Init();
  val = 0.0f;
}

void StatVal::InitLinks() {
  DataItem::InitLinks();
  taBase::Own(stopcrit, this);
}

void StatVal::Copy_(const StatVal& cp) {
  val = cp.val;
  str_val = cp.str_val;
  stopcrit = cp.stopcrit;
}

void StatVal_List::InitStat(float value) {
  int i;
  for(i=0; i<size; i++)
    ((StatVal*)FastEl(i))->InitStat(value);
}

void StatVal_List::Init() {
  int i;
  for(i=0; i<size; i++)
    ((StatVal*)FastEl(i))->Init();
}

bool StatVal_List::Crit() {
  int i;
  for(i=0; i<size; i++) {
    if(((StatVal*)FastEl(i))->Crit())
      return true;
  }
  return false;
}

void StatVal_List::NameStatVals(const char* nm, const char* opts, bool is_string) {
  if(size <= 0)    return;
  StatVal* sv = (StatVal*)FastEl(0);
  if(is_string)
    sv->SetStringVecNm(nm, size);
  else
    sv->SetFloatVecNm(nm, size);
  sv->AddDispOption(opts);
  int i;
  for(i=1; i<size; i++) {
    StatVal* sv = (StatVal*)FastEl(i);
    if(is_string)
      sv->SetStringName(nm);
    else
      sv->SetName(nm);
    sv->AddDispOption(opts);
  }
}

bool StatVal_List::HasStopCrit() {
  int i;
  for(i=0; i<size; i++) {
    StatVal* sv = (StatVal*)FastEl(i);
    if(sv->stopcrit.flag)
      return true;
  }
  return false;
}

void StatVal_List::GenLogData(LogData* ld) {
  int i;
  for(i=0; i<size; i++) {
    StatVal* sv = (StatVal*)FastEl(i);
    if(sv->is_string)
      ld->AddString(sv, sv->str_val);
    else
      ld->AddFloat(sv, sv->val);
  }
}

//////////////////////////
//  	StatValAgg     	//
//////////////////////////

void StatValAgg::Initialize() {
}

void StatValAgg::Destroy() {
}

void StatValAgg::NewCopyAgg(StatVal* to, float fm) {
  if(owner == NULL)	return;
  Stat* ownr = (Stat*)owner;
  StatVal* sv;
  if(ownr->copy_vals.size > ownr->n_copy_vals) {
    sv = (StatVal*)ownr->copy_vals.FastEl(ownr->n_copy_vals);
    if(sv->name.length() == 0)
      sv->name = String("cpy_") + to->name + "_" + String(ownr->n_copy_vals);
  }
  else {
    sv = (StatVal*)to->Clone();
    sv->name = to->name;
    if(sv->vec_n > 0) {		// get rid of any vector info..
      sv->vec_n = 0;
      if(sv->name.contains('>'))
	sv->name = sv->name.after('>', -1);
    }
    sv->name = String("cpy_") + sv->name + "_" + String(ownr->n_copy_vals);
    ownr->copy_vals.Add(sv);
  }
  sv->val = fm;
  ownr->n_copy_vals++;
}

void StatValAgg::ComputeAgg(StatVal* to, float fm) {
  if(op == COPY) {
    NewCopyAgg(to, fm);
    IncUpdt();
  }
  else {
    ComputeAgg(to->val, fm);
  }
}

void StatValAgg::ComputeAggNoUpdt(StatVal* to, float fm) {
  if(op == COPY) {
    NewCopyAgg(to, fm);
  }
  else {
    ComputeAggNoUpdt(to->val, fm);
  }
}

void StatValAgg::ComputeAggsNoUpdt(StatVal_List* to, StatVal_List* fm) {
  if(fm->size <= 0)
    return;
  to->EnforceSize(fm->size);
  int j;
  for(j=0; j<fm->size; j++) {
    ComputeAggNoUpdt(((StatVal*)to->FastEl(j)), ((StatVal*)fm->FastEl(j))->val);
  }
}

void StatValAgg::ComputeAggs(StatVal_List* to, StatVal_List* fm) {
  ComputeAggsNoUpdt(to,fm);
  IncUpdt();
}

// this assumes NoUpdt
void StatValAgg::AggFromCopy(StatVal_List* fm) {
  if(owner == NULL)	return;
  Stat* ownr = (Stat*)owner;

  if(op == COPY) {		// copying a copy, the list gets bigger..
    int j;
    for(j=0; j<fm->size; j++) {
      // use the from variable as a template for creating new copy statvals..
      ComputeAggNoUpdt(((StatVal*)fm->FastEl(j)), ((StatVal*)fm->FastEl(j))->val);
    }
  }
  else {
    int j;
    for(j=0; j<fm->size; j++) {
      StatVal* fmsv = (StatVal*)fm->FastEl(j);
      StatVal* sv;
      if(ownr->copy_vals.size > j)
	sv = (StatVal*)ownr->copy_vals.FastEl(j);
      else {
	sv = (StatVal*)fmsv->Clone();
	ownr->copy_vals.Add(sv);
      }
      ComputeAggNoUpdt(sv, fmsv->val);
    }
    ownr->n_copy_vals = fm->size;
  }
}
    

//////////////////////////
//  	AggStat     	//
//////////////////////////

void AggStat::Initialize() {
  from = NULL;
  real_stat = NULL;
  type_safe = true;
}

void AggStat::CutLinks() {
  taBase::DelPointer((TAPtr*)&from);
  StatValAgg::CutLinks();
}

void AggStat::Copy_(const AggStat& cp) {
  taBase::SetPointer((TAPtr*)&from,cp.from);
  type_safe = cp.type_safe;
}

void AggStat::UpdateAfterEdit() {
  StatValAgg::UpdateAfterEdit();

  if(owner == NULL)
    return;

  if(from != NULL) {
    if(type_safe && !owner->InheritsFrom(from->GetTypeDef())) {
      taMisc::Error("Incompatible Aggregation of type:",from->GetTypeDef()->name,
	       "into type:",owner->GetTypeDef()->name);
      from = NULL;
    }
  }
  FindRealStat();
  owner->UpdateAfterEdit();
}

void AggStat::SetFrom(Stat* frm) {
  taBase::SetPointer((TAPtr*)&from, frm);
}

void AggStat::FindRealStat() {
  if(owner == NULL) {
    real_stat = NULL;
    return;
  }

  Stat* tmp = (Stat*)owner;
  while ((tmp->time_agg.from != NULL) && (tmp->time_agg.from != tmp))
    tmp = tmp->time_agg.from;

  real_stat = tmp;
}

String AggStat::AppendAggName(const char* nm) const {
  String rval = nm;
  if(from == NULL)
    return rval;
  rval += ":";
  rval += GetAggName();
  return rval;
}

String AggStat::PrependAggName(const char* nm) const {
  if(from == NULL)
    return (String)nm;
  String rval = GetAggName();
  rval += "_";
  rval += nm;
  return rval;
}


//////////////////////////
//  	Stat     	//
//////////////////////////

void Stat::Initialize() {
  own_proc = NULL;
  has_stop_crit = false;
  n_copy_vals = 0;
  loop_init = true;
  log_stat = true;
  layer = NULL;
}

void Stat::InitLinks() {
  Process::InitLinks();
  own_proc = GET_MY_OWNER(SchedProcess);
  taBase::Own(time_agg, this);
  taBase::Own(net_agg, this);
  taBase::Own(copy_vals, this);
  // own all the stats in the object
  // just in this one case..
  TypeDef*   td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++){
    MemberDef* md=td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_StatVal))
      taBase::Own((StatVal*)md->GetOff(this), this);
    else if(md->type->InheritsFrom(&TA_StatVal_List))
      taBase::Own((StatVal_List*) md->GetOff(this), this);
  }
  NameStatVals();
  Init();
}

void Stat::CutLinks() {
  DeleteAggregates();
  copy_vals.CutLinks();
  time_agg.CutLinks();
  net_agg.CutLinks();
  taBase::DelPointer((TAPtr*)&layer);
  own_proc = NULL;
  Process::CutLinks();
}

void Stat::Copy_(const Stat& cp) {
  has_stop_crit = cp.has_stop_crit;
  n_copy_vals = cp.n_copy_vals;
  log_stat = cp.log_stat;
  time_agg = cp.time_agg;
  net_agg = cp.net_agg;
  copy_vals = cp.copy_vals;
  taBase::SetPointer((TAPtr*)&layer, cp.layer);
}

void Stat::NameStatVals() {
  Stat* rstat = time_agg.real_stat;
  String misc_nm;
  if((rstat != NULL) && (rstat->layer != NULL)) {
    misc_nm = rstat->layer->name;
    if(misc_nm.length() > 4)
      misc_nm = misc_nm.before(4);
  }
  // if non-copy aggregator of copy_vals, get names from real-stat
  if((time_agg.op != Aggregate::COPY) && (time_agg.from != NULL)
     && (rstat != NULL) && (copy_vals.size > 0) && (rstat->copy_vals.size > 0))
  {
    int mx = MIN(rstat->copy_vals.size, copy_vals.size);
    int i;
    for(i=0; i<mx; i++) {
      StatVal* sv = (StatVal*)copy_vals.FastEl(i);
      StatVal* rsv = (StatVal*)rstat->copy_vals.FastEl(i);
      String nm = rsv->name;
      if((i==0) && (rsv->vec_n > 0) && !nm.empty() && (nm[0] == '<'))
	nm = nm.after('>');
      nm = time_agg.PrependAggName(nm);
      if((i==0) && (copy_vals.size > 1)) // only for vec > 1
	sv->SetFloatVecNm(nm, copy_vals.size);
      else {
	sv->name = nm;
	sv->vec_n = 0;
      }
    }
  }
  TypeDef* td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++){
    MemberDef* md=td->members.FastEl(i);
    String nm;
    if(!misc_nm.empty())
      nm = misc_nm + "_" + md->name;
    else
      nm = md->name;
    String ag_nm = time_agg.PrependAggName(net_agg.PrependAggName(nm));
    if(md->type->InheritsFrom(&TA_StatVal)) {
      StatVal* sv = (StatVal*)md->GetOff(this);
      sv->SetName(ag_nm);
    }
    else if(md->type->InheritsFrom(&TA_StatVal_List)) {
      StatVal_List* svl = (StatVal_List*)md->GetOff(this);
      svl->NameStatVals(ag_nm, "", false);
    }
  }
}

bool Stat::HasStopCrit() {
  if((n_copy_vals > 0) && copy_vals.HasStopCrit())
    return true;
  TypeDef* td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++){
    MemberDef* md = td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_StatVal)) {
      if(((StatVal*) md->GetOff(this))->stopcrit.flag)
	return true;
    }
    else if(md->type->InheritsFrom(&TA_StatVal_List)) {
      if(((StatVal_List*) md->GetOff(this))->HasStopCrit())
	return true;
    }
  }
  return false;
}

void Stat::UpdateAfterEdit() {
  Process::UpdateAfterEdit();
  time_agg.FindRealStat();
  Stat* rstat = time_agg.real_stat;
  if((time_agg.from != NULL) && (rstat != NULL)) { // if aggregator
    net_agg = rstat->net_agg;	// always copy stuff from real stat
    taBase::SetPointer((TAPtr*)&layer, rstat->layer);
    if(time_agg.op != Aggregate::COPY) {
      copy_vals.EnforceSize(rstat->copy_vals.size); // always have same size as real stat
      n_copy_vals = copy_vals.size;
    }
  }
  else if(net_agg.op != Aggregate::COPY) {
    copy_vals.Reset();
    n_copy_vals = 0;
  }
  NameStatVals();
  has_stop_crit = HasStopCrit();
  if((own_proc != NULL) && name.contains(GetTypeDef()->name)) {
    String nw_nm = own_proc->name + "_" + GetTypeDef()->name;
    name = time_agg.PrependAggName(net_agg.PrependAggName(nw_nm));
    if(layer != NULL)
      name += String("_") + layer->name;
    else if((time_agg.real_stat != NULL) && (time_agg.real_stat->layer != NULL))
      name += String("_") + time_agg.real_stat->layer->name;
  }
  UpdateAggregates();		// update our aggreagators too!
}

float Stat::InitStatVal() {
  if(time_agg.from != NULL)
    return time_agg.InitAggVal();
  return net_agg.InitAggVal();
}

void Stat::InitStat_impl() {
  net_agg.Init();  copy_vals.InitStat(InitStatVal());
}  

// it is suggested that this general-purpose function be replaced by one
// that specifically calls InitStat() on StatVal objects directly and:
// InitStat_impl();

void Stat::InitStat() {
  float init_val = InitStatVal();
  TypeDef*   td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++) {
    MemberDef* md=td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_StatVal))
      ((StatVal*) md->GetOff(this))->InitStat(init_val);
    else if(md->type->InheritsFrom(&TA_StatVal_List))
      ((StatVal_List*) md->GetOff(this))->InitStat(init_val);
  }
  InitStat_impl();
}

void Stat::Init_impl() {
  time_agg.Init();  InitStat();	n_copy_vals = 0;
}  

// it is suggested that these general-purpose functions
// (Init and Crit) be replaced by specific ones

// Init() specifically calls Init() on StatVal objects directly, and:
// Init_impl();

void Stat::Init() {
  TypeDef*   td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++){
    MemberDef* md = td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_StatVal)) 
      ((StatVal*) md->GetOff(this))->Init();
    else if(md->type->InheritsFrom(&TA_StatVal_List)) 
      ((StatVal_List*) md->GetOff(this))->Init();
  }
  Init_impl();
}  

// Crit() checks has_stop_crit flag and if true, calls Crit() on statvals directly
// can replace Crit() with first two if statements plus direct call on statval

bool Stat::Crit() {
  if(!has_stop_crit)    return false;
  if(n_copy_vals > 0)   return copy_vals.Crit();

  TypeDef*   td = GetTypeDef();
  int i;
  for(i=TA_Stat.members.size; i<td->members.size;i++){
    MemberDef* md = td->members.FastEl(i);
    if(md->type->InheritsFrom(&TA_StatVal)) {
      if(((StatVal*) md->GetOff(this))->Crit())
	return true;
    }
    else if(md->type->InheritsFrom(&TA_StatVal_List)) {
      if(((StatVal_List*) md->GetOff(this))->Crit())
	return true;
    }
  }
  return false;
}  

void Stat::C_Code() {
  if(time_agg.from != NULL)
    ComputeAggregates();
  else
    Network_Run();
}

void Stat::Network_Run() {
  if(network == NULL)
    return;
  Network_Init();
  Layer_Run();
  Network_Stat();
}

bool Stat::CheckLayerInNet() {
  if(layer == NULL) return true;
  if(layer->own_net == network)
    return true;

  taMisc::Error("*** Stat:", name, "layer:", layer->name,
		 "is not in current network, finding one of same name...");
  Layer* nw_lay = (Layer*)network->layers.FindName(layer->name);
  if(nw_lay == NULL) {
    taMisc::Error("==> could not find layer with same name in current network, aborting");
    return false;
  }
  taBase::SetPointer((TAPtr*)&layer, nw_lay);
  return true;
}
 
void Stat::Layer_Run() {
  if(layer != NULL) {
    if(!CheckLayerInNet()) return;
    if(layer->lesion)      return;
    Layer_Init(layer);
    Unit_Run(layer);
    Layer_Stat(layer);
    return;
  }
  Layer* lay;
  taLeafItr i;
  FOR_ITR_EL(Layer, lay, network->layers., i) {
    if(lay->lesion)
      continue;
    Layer_Init(lay);
    Unit_Run(lay);
    Layer_Stat(lay);
  }
}
void Stat::Unit_Run(Layer* lay) {
  Unit* unit;
  taLeafItr i;
  FOR_ITR_EL(Unit, unit, lay->units., i) {
    Unit_Init(unit);
    RecvCon_Run(unit);
    Unit_Stat(unit);
  }
}

void Stat::RecvCon_Run(Unit* unit) {
  Con_Group* cg;
  int i;
  FOR_ITR_GP(Con_Group, cg, unit->recv., i) {
    int j;
    for(j=0; j<cg->size; j++)
      Con_Stat(unit, cg->Cn(j), cg->Un(j));
  }
}

void Stat::SendCon_Run(Unit* unit) {
  Con_Group* cg;
  int i;
  FOR_ITR_GP(Con_Group, cg, unit->send., i) {
    int j;
    for(j=0; j<cg->size; j++)
      Con_Stat(cg->Un(j), cg->Cn(j), unit);
  }
}

LogData& Stat::GenLogData(LogData* ld) {
  if(ld == NULL)
    ld = &log_data;
  if(!log_stat)  return *ld;

  if(n_copy_vals > 0) {
    copy_vals.EnforceSize(n_copy_vals);
    if(copy_vals.size > 1) {
      StatVal* sv = (StatVal*)copy_vals.FastEl(0);
      if(sv->vec_n != copy_vals.size) {
	String nm = sv->name;
	if(nm.contains('>'))
	  nm = nm.after('>', -1);
	if(nm.empty())
	  nm = ((StatVal*)copy_vals.FastEl(1))->name;
	sv->SetFloatVecNm(nm, copy_vals.size);
      }
    }
    copy_vals.GenLogData(ld);
  }
  else {
    TypeDef* td = GetTypeDef();
    int i;
    for(i=TA_Stat.members.size; i<td->members.size;i++){
      MemberDef* md=td->members.FastEl(i);
      if(md->type->InheritsFrom(&TA_StatVal)) {
	StatVal* sv = (StatVal*)md->GetOff(this);
	if(sv->is_string)
	  ld->AddString(sv, sv->str_val);
	else
	  ld->AddFloat(sv, sv->val);
      }
      else if(md->type->InheritsFrom(&TA_StatVal_List)) {
	((StatVal_List*)md->GetOff(this))->GenLogData(ld);
      }
    }
  }
  return *ld;
}
  
void Stat::ComputeAggregates() {
  if(time_agg.from == NULL)
    return;

  if(time_agg.from->n_copy_vals > 0) {
    time_agg.AggFromCopy(&(time_agg.from->copy_vals));
  }
  else {
    TypeDef* td = time_agg.from->GetTypeDef();
    int i;
    for(i=TA_Stat.members.size; i< td->members.size; i++) {
      MemberDef* md = td->members.FastEl(i);

      if(md->type->InheritsFrom(&TA_StatVal)) {
	StatVal* fm = (StatVal*) md->GetOff((void*)time_agg.from);
	StatVal* to = (StatVal*) md->GetOff((void*)this);
	time_agg.ComputeAggNoUpdt(to, fm->val);
      }
      else if(md->type->InheritsFrom(&TA_StatVal_List)) {
	StatVal_List* fm = (StatVal_List *) md->GetOff((void*)time_agg.from);
	StatVal_List* to = (StatVal_List *) md->GetOff((void*)this);
	time_agg.ComputeAggsNoUpdt(to, fm);
      }
    }
  }
  time_agg.IncUpdt();		// only inc the updt once at the end..
}

void Stat::CreateAggregates(Aggregate::Operator default_op) {
  if(own_proc == NULL)
    return;

  SchedProcess* sproc= (SchedProcess *) own_proc->super_proc;
  Stat*		stat_to_agg = this;

  while((sproc != NULL) && (sproc->InheritsFrom(&TA_SchedProcess))) {
    String_PArray& typeopts = sproc->GetTypeDef()->opts;
    int use_op = default_op;
    if((stat_to_agg != this) || (default_op == (Aggregate::Operator)-1)) {
      int opt;
      if((opt = typeopts.FindContains("AGGOP_")) >= 0) {
	String op = typeopts.FastEl(opt).after("AGGOP_");
	EnumDef* lop;
	if((lop = (TA_AggStat.FindEnum((char*)op))) != NULL) {
	  use_op = lop->enum_no;
	}
      }
      else {
	use_op = AggStat::LAST;
      }
    }
    Stat* nag = (Stat*)stat_to_agg->Clone(); // clone original one
    sproc->loop_stats.Add(nag);
    nag->time_agg.op = (AggStat::Operator)use_op;
    taBase::SetPointer((TAPtr*)&(nag->time_agg.from), stat_to_agg);
    nag->time_agg.UpdateAfterEdit();
    nag->UpdateAfterEdit();
    sproc = (SchedProcess *) sproc->super_proc;
    stat_to_agg = nag;
  }
  UpdateAfterEdit();
  winbMisc::DelayedMenuUpdate(own_proc);
}

void Stat::DeleteAggregates() {
  if(own_proc == NULL) return;
  SchedProcess* sproc= (SchedProcess *) own_proc->super_proc;
  Stat*		stat_to_agg = this;

  while((sproc != NULL) && (sproc->InheritsFrom(&TA_SchedProcess))) {
    Stat* next_stat_to_agg = NULL;
    int i;
    for(i=sproc->loop_stats.size-1; i>=0; i--) {
      Stat* ast = (Stat*)sproc->loop_stats.FastEl(i);
      if(ast->time_agg.from == stat_to_agg) {
	if(next_stat_to_agg == NULL) // get the first one (most likely to be big agg)
	  next_stat_to_agg = ast;
	else
	  ast->DeleteAggregates(); // be thorough anyway (a little recursion can't hurt)
	sproc->loop_stats.Remove(i);
      }
    }
    if(next_stat_to_agg != NULL)
      stat_to_agg = next_stat_to_agg;
    sproc = (SchedProcess *) sproc->super_proc;
  }
}

void Stat::UpdateAggregates() {
  Stat* aggr = FindAggregator();
  if(aggr != NULL) {
    aggr->UpdateAfterEdit();
//    aggr->UpdateAggregates();
  }
}

Stat* Stat::FindAggregator() {
  if(own_proc == NULL) return NULL;
  SchedProcess* sproc= (SchedProcess *) own_proc->super_proc;
  Stat*		stat_to_agg = this;

  while((sproc != NULL) && (sproc->InheritsFrom(&TA_SchedProcess))) {
    int i;
    for(i=sproc->loop_stats.size-1; i>=0; i--) {
      Stat* ast = (Stat*)sproc->loop_stats.FastEl(i);
      if(ast->time_agg.from == stat_to_agg)
	return ast;
    }
    sproc = (SchedProcess *) sproc->super_proc;
  }
  return NULL;
}


//////////////////////////
// 	SE_Stat		//
//////////////////////////

void SE_Stat::Initialize() {
  tolerance = 0.0f;
}

void SE_Stat::NameStatVals() {
  Stat::NameStatVals();
  se.AddDispOption("MIN=0, ");
  se.AddDispOption("TEXT, ");
}

void SE_Stat::InitStat() {
  float init_val = InitStatVal();
  se.InitStat(init_val);
  InitStat_impl();
}

void SE_Stat::Init() {
  se.Init();
  Init_impl();
}

bool SE_Stat::Crit() {
  if(!has_stop_crit)    return false;
  if(n_copy_vals > 0)   return copy_vals.Crit();
  return se.Crit();
}

void SE_Stat::Network_Init() {
  InitStat();
}

void SE_Stat::Layer_Run() {
  if(layer != NULL) {
    if(!CheckLayerInNet()) return;
    if(layer->lesion)      return;
    Layer_Init(layer);
    Unit_Run(layer);
    Layer_Stat(layer);
    return;
  }
  Layer *lay;
  taLeafItr i;
  FOR_ITR_EL(Layer, lay, network->layers., i) {
    // only target layers need to be checked for error
    if(lay->lesion || !((lay->ext_flag & Unit::TARG) || (lay->ext_flag & Unit::COMP)))
      continue;
    Layer_Init(lay);
    Unit_Run(lay);
    Layer_Stat(lay);
  }
}

void SE_Stat::Unit_Stat(Unit* unit) {
  if(!((unit->ext_flag & Unit::TARG) || (unit->ext_flag & Unit::COMP)))
    return;
  float tmp = fabsf(unit->act - unit->targ);
  if(tmp >= tolerance)
    tmp *= tmp;
  else
    tmp = 0.0f;
  net_agg.ComputeAgg(&se, tmp);
}

///////////////////////////
// 	Monitor Stat	 //
///////////////////////////

void MonitorStat::Initialize(){
  net_agg.op = Aggregate::COPY;
  variable = "act";
}

void MonitorStat::InitLinks(){
  Stat::InitLinks();
  taBase::Own(objects,this);
  taBase::Own(ptrs,this);
  objects.SetBaseType(&TA_taBase);
  ptrs.SetBaseType(&TA_taBase);
  taBase::Own(pre_proc_1,this);
  taBase::Own(pre_proc_2,this);
  taBase::Own(pre_proc_3,this);
}

void MonitorStat::CutLinks() {
  MonitorStat* rstat = (MonitorStat*)time_agg.real_stat;
  int i;
  for(i=0;i<rstat->objects.size;i++) {
    TAPtr obj = rstat->objects.FastEl(i);
    if(obj == NULL)
      continue;
    Network* net = GET_OWNER(obj, Network);
    if(net != NULL)
      winbMisc::DelayedMenuUpdate(net);
  }
  objects.RemoveAll();
  ptrs.RemoveAll();
  Stat::CutLinks();
}

void MonitorStat::Copy_(const MonitorStat& cp) {
  mon = cp.mon;
  objects.BorrowUnique(cp.objects);	// borrow because its a link group
  variable = cp.variable;
  pre_proc_1 = cp.pre_proc_1;
  pre_proc_2 = cp.pre_proc_2;
  pre_proc_3 = cp.pre_proc_3;
}

void MonitorStat::UpdateAfterEdit() {
  Stat::UpdateAfterEdit();
  if((owner == NULL) || (own_proc == NULL) || (time_agg.real_stat == NULL))    return;
  MonitorStat* rstat = (MonitorStat*)time_agg.real_stat;
  if(rstat->variable.empty())
    return;
  if(time_agg.from != NULL) {
    variable = rstat->variable;
    objects.RemoveAll();
    ptrs.RemoveAll();
    members.RemoveAll();
  }
  else {
    ScanObjects();
  }
  // add variable and object name, if relevant
  if(name.contains(GetTypeDef()->name)) {
    name += String("_") + variable;
    if(rstat->objects.size > 0) {
      String objnm = rstat->objects.FastEl(0)->GetName();
      if(!objnm.empty())
	name += String("_") + objnm;
    }
  }
}

void MonitorStat::NameStatVals() {
  Stat::NameStatVals();

  String nm = variable;
  if(nm.length() > 4)
    nm = nm.before(4);
  MonitorStat* rstat = (MonitorStat*)time_agg.real_stat;
  if((rstat != NULL) && (rstat->objects.size > 0)) {
    String objnm = rstat->objects.FastEl(0)->GetName();
    if(objnm.length() > 4)
      objnm = objnm.before(4);
    if(!objnm.empty())
      nm += String("_") + objnm;
  }
  mon.name.gsub("mon", nm);	// replace mon with name!
}
  
void MonitorStat::InitStat() {
  mon.InitStat(InitStatVal());
  InitStat_impl();
}

void MonitorStat::Init() {
  mon.Init();
  Init_impl();
}

bool MonitorStat::Crit() {
  if(!has_stop_crit)    return false;
  if(n_copy_vals > 0)   return copy_vals.Crit();
  return mon.Crit();
}

void MonitorStat::ScanConGroup(Con_Group* mcg, char* varname, Projection* p) {
  Con_Group* cg;
  int i;
  MemberDef* md;
  FOR_ITR_GP(Con_Group, cg, mcg->, i) {
    if((p != NULL) && (cg->prjn != p)) continue;
    md = cg->el_typ->members.FindNameR(varname);
    if(md == NULL) continue;
    String unitname;
    if(net_agg.op==Aggregate::COPY) {
      Unit* u = GET_OWNER(cg,Unit);
      if(u == NULL) continue;
      unitname = u->name;
      if(unitname.empty() && (u->owner != NULL)) {
	int index = ((Unit_Group *) u->owner)->Find(u);
	unitname = String("[") + String(index) + "]";
      }
      unitname = unitname + "[" + String(i) + "]";
    }
    int j;
    for(j=0; j<cg->size; j++) {
      Connection* c = (Connection *) cg->Cn(j);
      ptrs.Link(c);
      members.Add(md);
      if(net_agg.op==Aggregate::COPY) {
	StatVal* sv = (StatVal*)copy_vals.New(1);
	sv->name = unitname + "[" + String(j) + "]." + varname;
      }
    }
  }
}

void MonitorStat::ScanUnit(Unit* u, Projection* p) {
  TAPtr ths = u;
  void* temp;
  MemberDef* md = NULL;
  String varname = variable;
  if(variable.contains('.')) {
    varname = variable.before('.');
    // for selected projections:
    if(varname=="r") varname = "recv"; 
    else if(varname=="s") varname = "send";
    md = u->FindMembeR(varname,temp);
    if((md == NULL) || (temp == NULL)) return;
    if(md->type->ptr == 1) {
      ths = *((TAPtr*)temp);
      if(ths == NULL) return;
    }
    else
      ths = (TAPtr) temp;
    varname = variable.after('.');
    if(ths->InheritsFrom(&TA_Con_Group)) {
      ScanConGroup((Con_Group *) ths,varname,p);
      return;
    }
    // otherwise it must be bias or someother sub object
  }
  if(ths == NULL) return;
  md = ths->FindMember((const char*)varname);
  if(md == NULL) return;
  members.Add(md);
  ptrs.Link(ths);
  if(net_agg.op == Aggregate::COPY) {
    StatVal* sv = (StatVal*)copy_vals.New(1);
    String valname = u->name;
    if(valname.empty() && (u->owner != NULL)) {
      int index = ((Unit_Group *) u->owner)->Find(u);
      valname = String("[") + String(index) + "]";
    }
    else if(valname.length() > 4)
      valname = valname.before(4);
    valname = valname + "." + variable;
    sv->name = valname;
  }
}

void MonitorStat::ScanProjection(Projection* p) {
  int temp=0;
  MemberDef* md = p->FindMember((char *) variable,temp);
  if(md != NULL) {
    StatVal* sv = (StatVal*)copy_vals.New(1);
    sv->name = p->name;
    if(sv->name.length() > 4)
      sv->name = sv->name.before(4);
    sv->name += String(".") + variable;
    ptrs.Link(p);
    members.Add(md);
    return;
  }
  Layer* lay = NULL;
  if(variable.before('.') == "r") lay = p->layer;
  else if(variable.before('.') == "s") lay = p->from;
  if(lay == NULL) {
    taMisc::Error("MonitorStat Projection does not have layer's set or",
		   "selected variable does not apply to connections");
    return;
  }
  taLeafItr i;
  Unit* u;
  FOR_ITR_EL(Unit, u, lay->units., i)
    ScanUnit(u, p);
}

void MonitorStat::ScanLayer(Layer* lay) {
  TAPtr ths = lay;
  MemberDef* md = NULL;
  String varname = variable;
  if(variable.contains('.')) {
    varname = variable.before('.');
    md = lay->FindMember((const char*)varname);
    if(md != NULL) {
      if(md->type->ptr == 1)
	ths = *((TAPtr*)md->GetOff((void*)lay));
      else
	ths = (TAPtr) md->GetOff((void*)lay);
      varname = variable.after('.');
    }
    else			// variable not there, go to units
      ths = NULL;
  }
  if(ths != NULL) {
    // now search from 'ths'
    md = ths->FindMember((const char*)varname);
    if(md != NULL) {
      StatVal* sv = (StatVal*)copy_vals.New(1);
      sv->name = lay->name;
      if(sv->name.length() > 4)
	sv->name = sv->name.before(4);
      sv->name += String(".") + variable;
      ptrs.Link(ths);
      members.Add(md);
      return;
    }
  }
  // the default is to scan the units for the variable
  taLeafItr i;
  Unit* u;
  FOR_ITR_EL(Unit, u, lay->units., i)
    ScanUnit(u);
  StatVal* sv = (StatVal *) copy_vals[0];
  if(sv != NULL) {		// if not copying, this will be null..
    String xgeom = String("GEOM_X=") + String(lay->geom.x) + ", ";
    String ygeom = String("GEOM_Y=") + String(lay->geom.y) + ", ";
    sv->AddDispOption(xgeom.chars());
    sv->AddDispOption(ygeom.chars());
  }
}

void MonitorStat::ScanUnitGroup(Unit_Group* ug) {
  TAPtr ths = ug;
  MemberDef* md = NULL;
  String varname = variable;
  if(variable.contains('.')) {
    varname = variable.before('.');
    md = ug->FindMember((const char*)varname);
    if(md != NULL) {
      if(md->type->ptr == 1)
	ths = *((TAPtr*)md->GetOff((void*)ug));
      else
	ths = (TAPtr) md->GetOff((void*)ug);
      varname = variable.after('.');
    }
    else			// variable not there, go to units
      ths = NULL;
  }
  if(ths != NULL) {
    // now search from 'ths'
    md = ths->FindMember((const char*)varname);
    if(md != NULL) {
      StatVal* sv = (StatVal*)copy_vals.New(1);
      sv->name = ug->name;
      if(sv->name.length() > 4)
	sv->name = sv->name.before(4);
      sv->name += String(".") + variable;
      ptrs.Link(ths);
      members.Add(md);
      return;
    }
  }
  // the default is to scan the units for the variable
  taLeafItr i;
  Unit* u;
  FOR_ITR_EL(Unit, u, ug->, i)
    ScanUnit(u);
  StatVal* sv = (StatVal *) copy_vals[0];
  if(sv != NULL) {		// if not copying, this will be null..
    String xgeom = String("GEOM_X=") + String(ug->geom.x) + ", ";
    String ygeom = String("GEOM_Y=") + String(ug->geom.y) + ", ";
    sv->AddDispOption(xgeom.chars());
    sv->AddDispOption(ygeom.chars());
  }
}

void MonitorStat::ScanObjects() {
  if((time_agg.real_stat == NULL) || (time_agg.from != NULL))
    return;
  MonitorStat* rstat = (MonitorStat*)time_agg.real_stat;
  copy_vals.RemoveAll();
  tabMisc::NotifyEdits(&copy_vals);
  ptrs.RemoveAll();
  members.RemoveAll();
  int i;
  for(i=0;i<rstat->objects.size;i++){
    TAPtr obj = rstat->objects.FastEl(i);
    if(obj == NULL)
      continue;
    if(obj->InheritsFrom(&TA_Unit)) ScanUnit((Unit*) obj);
    else if(obj->InheritsFrom(&TA_Layer)) ScanLayer((Layer*) obj);
    else if(obj->InheritsFrom(&TA_Unit_Group)) ScanUnitGroup((Unit_Group*) obj);
    else if(obj->InheritsFrom(&TA_Projection))
      ScanProjection((Projection*) obj);
    else {			// could be any type of object 
      int temp = 0;
      MemberDef* md = obj->FindMember((char *) variable,temp);
      if(md != NULL) {
	StatVal* sv = (StatVal*)copy_vals.New(1);
	sv->name = obj->GetName() + "." + variable;
	ptrs.Link(obj);
	members.Add(md);
      }
    }
  }
  if(copy_vals.size > 1) {
    StatVal* sv = (StatVal*)copy_vals.FastEl(0);
    String nm = sv->name;
    sv->SetFloatVecNm(nm, copy_vals.size);
    n_copy_vals = copy_vals.size;
  }
}
      
void MonitorStat::Network_Init() {
  InitStat();
}

void MonitorStat::Network_Stat() {
  if((time_agg.real_stat == NULL) || (variable.empty()))
    return;
  int i;
  for(i=0;i<members.size;i++) {
    MemberDef* md = members.FastEl(i);
    float mbval = 0.0f;
    if((md->type == &TA_float) ||
       (md->type->InheritsFrom(&TA_float)))
       mbval = *((float*)md->GetOff((void*)ptrs.FastEl(i)));
    else if(md->type->InheritsFrom(&TA_int)){
      mbval = (float) *((int *) md->GetOff((void*)ptrs.FastEl(i)));
    }
    // pre-process..
    mbval = pre_proc_3.Evaluate(pre_proc_2.Evaluate(pre_proc_1.Evaluate(mbval)));
    net_agg.ComputeAgg(&mon, mbval); // automatically puts copy vals in copy..
  }
}

void MonitorStat::SetVariable(const char* varnm){
  variable = varnm;
  UpdateAfterEdit();
}

void MonitorStat::SetObject(TAPtr obj){
  objects.Reset();
  objects.Link(obj);
  UpdateAfterEdit();
}

void MonitorStat::AddObject(TAPtr obj){
  objects.Link(obj);
  UpdateAfterEdit();
}

